from collections import namedtuple
from domainpy.infrastructure.projectionsourced.recordmanager import (
    ProjectionRecordManager,
    Session
)

Operation = namedtuple('Operation', ('type', 'key', 'item'))

class MemoryProjectionRecordManager(ProjectionRecordManager):
    
    def __init__(self):
        self._heap = []
        
    def session(self):
         return MemorySession(self)
     
    def get_item(self, key, attributes):
        items = []

        for item in self._heap:
            if key.items() <= item.items():
                items.append({ key: item[key] for key in attributes })

        return tuple(items)
        
    def get_all_items(self, attributes):
        items = []

        for item in self._heap:
            items.append({ key: item[key] for key in attributes })

        return tuple(items)

class MemorySession(Session):
    
    def __init__(self, record_manager):
        self.record_manager = record_manager

        self._log = []
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args, **kwargs):
        self.rollback()
    
    def put_item(self, item):
        if item is None:
            raise TypeError('item cannot be none')
        
        self._log.append(
            Operation(
                type='put',
                key=None,
                item=item
            )
        )

    def update_item(self, key, item):
        self._log.append(
            Operation(
                type='update',
                key=key,
                item=item
            )
        )

    def delete_item(self, key):
        self._log.append(
            Operation(
                type='delete',
                key=key,
                item=None
            )
        )
    
    def commit(self):
        for operation in self._log:
            if(operation.type == 'put'):
                self._put_item(operation.item)
            elif (operation.type == 'update'):
                self._update_item(operation.key, operation.item)
            elif (operation.type == 'delete'):
                self._delete_item(operation.key)
    
    def rollback(self):
        pass

    def _put_item(self, item):
        self.record_manager._heap.append(item)

    def _update_item(self, key, item):
        items = (i for i in self.record_manager._heap if self._match(key, i))
        for i in items:
            if callable(item):
                i.update(item(i))
            else:
                i.update(item)

    def _delete_item(self, key):
        items = (i for i in self.record_manager._heap if self._match(key, i))
        for i in items:
            self.record_manager._heap.remove(i)

    def _match(self, key, item):
        return key.items() <= item.items()




from .command import ApplicationCommand
from .integration import IntegrationEvent

class tracer:

    def __call__(self, func):
        def wrapper(service, message):

            if isinstance(message, (ApplicationCommand, IntegrationEvent)):
                self.trace = [(service, message)]
            else:
                self.trace.append((service, message))

            return func(service, message)
        return wrapper

    def inject(self, func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs, trace=self.trace)

        return wrapper





class Registry:
    
    def __init__(self):
        self.components = {}

    def has(self, key):
        return key in self.components

    def assert_component(self, key):
        assert self.has(key)

    def put(self, key, value):
        self.components[key] = value

    def get(self, key):
        return self.components[key]
    




class EventRecordManager:
    
    def session(self):
        raise NotImplementedError(f'{self.__class__.__name__} must override session method')
        
    def find(self, stream_id):
        raise NotImplementedError(f'{self.__class__.__name__} must override find method')


class Session:
    
    def append(self, event_record):
        raise NotImplementedError(f'{self.__class__.__name__} must override append method')




from .mappers.integrationmapper import IntegrationMapper, IntegrationRecord
from .mappers.commandmapper import CommandMapper, CommandRecord
from .mappers.eventmapper import EventMapper, EventRecord

from .bus import Bus, Commutator
from .registry import Registry
from .subscriber import Subscriber

from .constructable import Constructable
from .dictable import Dictable



from .integration import IntegrationEvent
from .command import ApplicationCommand
from .handler import handler
from .tracer import tracer
from .projection import Projection
from .projector import projector
from .query import ApplicationQuery
from .service import ApplicationService




from domainpy.utils.bus import Bus


class MemoryBus(Bus):

    def __init__(self, publish_exceptions=[]):
        self._handlers = []
        self.publish_exceptions = tuple(publish_exceptions)

    def attach(self, handler):
        self._handlers.append(handler)

    def detach(self, handler):
        self._handlers.remove(handler)

    def publish(self, publishable):
        exceptions = []

        for handler in self._handlers:
            try:
                handler.__handle__(publishable)
            except Exception as e:
                if len(self.publish_exceptions) > 0 and isinstance(e, self.publish_exceptions):
                    exceptions.append(e)
                else:
                    raise e

        for e in exceptions:
            self.publish(e)
    



class HandlerNotFoundError(Exception):
    pass


class MessageSingleHandlerBroken(Exception):
    pass


class ExtractorMatchNotFound(Exception):
    pass
    

class MalformedMessageError(Exception):
    pass




from domainpy.utils.constructable import Constructable
from domainpy.utils.immutable import Immutable
from domainpy.utils.dictable import Dictable
from domainpy.utils.traceable import Traceable


class DomainEvent(Constructable, Immutable, Dictable, Traceable):
    
    def __init__(self, *args, **kwargs):
        self.__dict__.update({
            '__aggregate_id__': kwargs.pop('__aggregate_id__', None),
            '__number__': kwargs.pop('__number__', None),
            '__version__': kwargs.pop('__version__', None),
            '__timestamp__': kwargs.pop('__timestamp__', None),
            '__trace_id__': kwargs.pop('__trace_id__', Traceable.__trace_id__),
            '__message__': 'event'
        })
        
        super(DomainEvent, self).__init__(*args, **kwargs)
    
    def __eq__(self, other):
        if other is None:
            return False

        if self.__class__ != other.__class__:
            return False

        return (
            self.__aggregate_id__ == other.__aggregate_id__
            and self.__number__ == other.__number__
        )




class Constructable:
    
    def __init__(self, *args, **kwargs):
        if hasattr(self.__class__, '__annotations__'):
            annotations = self.__class__.__dict__['__annotations__']
            
            args_count = len(args)
            
            kwargs0 = {}
            
            current_arg = 0
            for k in annotations:
                expected_type = annotations[k]
                
                if current_arg <= args_count - 1:
                    value = args[current_arg]
                else:
                    try:
                        value = kwargs[k]
                    except KeyError:
                        raise TypeError(f'Missing argument {k} for '
                                    f'{self.__class__.__name__}')
                
                if hasattr(expected_type, '__origin__') and expected_type.__origin__ in (list, tuple):
                    (expected_type0,) = expected_type.__args__
                    for v in value:
                        if not isinstance(v, expected_type0):
                            raise TypeError(f'Bad type in {k}: Value {v} expected to be {expected_type0}')
                    
                    kwargs0[k] = value
                elif isinstance(value, expected_type):
                    kwargs0[k] = value
                else:
                    raise TypeError(f'Class {self.__class__.__name__}: {k} should be instance of {expected_type},'
                                    f' found {value.__class__.__name__} ({value})')
                
                current_arg = current_arg + 1
                
            self.__dict__.update(**kwargs0)



import boto3
import json

from domainpy.utils.bus import Bus


class EventBridgeBus(Bus):

    def __init__(self, source, mapper, region_name=None):
        self.source = source
        self.mapper = mapper

        self.cloudwatch_events = boto3.client('events', region_name=region_name)
        self.names = []

    def attach(self, handler):
        self.names.append(handler)

    def detach(self, handler):
        self.names.remove(handler)

    def publish(self, publishable):
        for name in self.names:
            self.cloudwatch_events.put_events(
                Entries=[
                    {
                        'Source': self.source,
                        'Detail': json.dumps(self.mapper.serialize(publishable)._asdict()),
                        'DetailType': publishable.__class__.__name__,
                        'EventBusName': name
                    }
                ]
            )
    



from functools import update_wrapper, partial

from domainpy.domain.model.event import DomainEvent
from domainpy.domain.exceptions import (
    MutatorNotFoundError,
    SingleMutatorBrokenError
)

class mutator:
    
    def __init__(self, func):
        update_wrapper(self, func)
        
        self.func = func
        
        self._events = dict()
        
    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)
        
    def __call__(self, aggregate, event: DomainEvent):
        if(event.__class__ not in self._events):
            # raise MutatorNotFoundError(event.__class__.__name__ + " in " + aggregate.__class__.__name__)
            return
        
        m = self._events[event.__class__]
        m(aggregate, event)
        
    def event(self, event_type: type):
        def inner_function(func):
            
            if event_type in self._events:
                raise SingleMutatorBrokenError(f'{event_type.__name__} mutator already registerd')
            
            self._events[event_type] = func
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function



from collections import namedtuple


CommandRecord = namedtuple(
    'CommandRecord', 
    ('trace_id', 'topic', 'version', 'timestamp', 'message', 'payload')
)


class CommandMapper:
    
    def __init__(self):
        self.map = dict()
        
    def register(self, cls):
        self.map[cls.__name__] = cls
        return cls

    def is_command(self, topic):
        return topic in self.map
    
    def serialize(self, command):
        if hasattr(command.__class__, '__annotations__'):
            attrs = command.__class__.__dict__['__annotations__']
            
            return CommandRecord(
                trace_id=command.__trace_id__,
                topic=command.__class__.__name__,
                version=command.__version__, # pylint: disable=maybe-no-member
                timestamp=command.__timestamp__, # pylint: disable=maybe-no-member
                message=command.__message__,
                payload=command.__to_dict__()
            )
        else:
            raise NotImplementedError(
                f'{command.__class__.__name__} should have annotations'
            )
    
    def deserialize(self, command_record: CommandRecord):
        command_class = self.map[command_record.topic]
        command = command_class.__from_dict__(command_record.payload)
        
        command.__dict__.update({
            '__trace_id__': command_record.trace_id,
            '__version__': command_record.version,
            '__timestamp__': command_record.timestamp,
            '__message__': command_record.message
        })
        
        return command
    



from domainpy.infrastructure.eventsourced.eventstream import EventStream


class EventStore:
    
    def __init__(self, event_mapper, record_manager, bus):
        self.event_mapper = event_mapper
        self.record_manager = record_manager
        self.bus = bus
    
    def store_events(self, stream):
        with self.record_manager.session() as session:
            for e in stream:
                session.append(
                    self.event_mapper.serialize(e)
                )
            
            session.commit()

            for e in stream:
                self.bus.publish(e)
        
    def get_events(self, stream_id: str):
        events = self.record_manager.find(
            stream_id=stream_id
        )
        
        stream = EventStream()
        for e in events:
            stream.append(
                self.event_mapper.deserialize(e)
            )
            
        return stream
        



from collections import namedtuple


IntegrationRecord = namedtuple(
    'IntegrationRecord', 
    ('trace_id', 'context', 'topic', 'resolve', 'version', 'timestamp', 'message', 'error', 'payload')
)


class IntegrationMapper:
    
    def __init__(self, context):
        self.context = context

        self.map = dict()
        
    def register(self, cls):
        self.map[cls.__name__] = cls
        return cls
    
    def is_integration(self, topic):
        return topic in self.map
    
    def serialize(self, event):
        if hasattr(event.__class__, '__annotations__'):
            attrs = event.__class__.__dict__['__annotations__']
            
            return IntegrationRecord(
                trace_id=event.__trace_id__, # pylint: disable=maybe-no-member
                context=self.context, # pylint: disable=maybe-no-member
                topic=event.__class__.__name__,
                resolve=event.__resolve__,
                version=event.__version__, # pylint: disable=maybe-no-member
                timestamp=event.__timestamp__, # pylint: disable=maybe-no-member
                message=event.__message__,
                error=event.__error__,
                payload=event.__to_dict__()
            )
        else:
            raise NotImplementedError(
                f'{event.__class__.__name__} should have annotations'
            )
    
    def deserialize(self, integration_record: IntegrationRecord):
        integration_class = self.map[integration_record.topic]
        integration = integration_class.__from_dict__(integration_record.payload)
        
        integration.__dict__.update({
            '__trace_id__': integration_record.trace_id,
            '__context__': integration_record.context,
            '__version__': integration_record.version,
            '__timestamp__': integration_record.timestamp,
            '__message__': integration_record.message,
            '__error__': integration_record.error
        })
        
        return integration
    


import pytest

from .example import ExampleValueObject

@pytest.fixture
def value_object_1():
    return ExampleValueObject('a')

@pytest.fixture
def value_object_2():
    return ExampleValueObject('a')

def test_value_object_equality(value_object_1, value_object_2):
    assert value_object_1 == value_object_2

def test_value_object_immutable(value_object_1):
    with pytest.raises(Exception):
        value_object_1.something = 'b'




from domainpy.domain.model.value_object import Identity

class DomainEntity:
    
    def __init__(self, id, aggregate):
        self.__id__ = id
        self.__aggregate__ = aggregate
        
    def __apply__(self, event):
        self.__aggregate__.__apply__(event)
    
    def __route__(self, event):
        self.mutate(event)
    
    def __eq__(self, other):
        if other is None:
            return False
        
        if isinstance(other, self.__class__):
            return self.__id__ == other.__id__
        elif isinstance(other, self.__id__.__class__):
            return self.__id__ == other
        else:
            return False
            
    def __repr__(self):
        return f'{self.__class__.__name__}(id={self.__id__})'
    
    def mutate(self, event):
        pass



from functools import update_wrapper, partial

from domainpy.application.command import ApplicationCommand
from domainpy.application.service import ApplicationService
from domainpy.application.exceptions import (
    HandlerNotFoundError,
    MessageSingleHandlerBroken
)
from domainpy.utils.traceable import Traceable


class handler:
    
    def __init__(self, func):
        update_wrapper(self, func)
        
        self.func = func
        
        self._handlers = dict()
        
    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)
        
    def __call__(self, service, message):
        if not self._has_any_handler(service, message):
            return

        if hasattr(message, '__trace_id__'):
            Traceable.__trace_id__ = message.__trace_id__
        else:
            Traceable.__trace_id__ = None

        results = [self.func(service, message)]
        
        handlers = self._handlers.get(message.__class__, set())
        for h in handlers:
            results.append(h(service, message))

        if(hasattr(service, '__partials__')):
            partials = service.__partials__.get(message.__class__, set())
            for h in set(partials):
                results.append(h(service, message))
                partials.remove(h)

        return results

    def _has_any_handler(self, service, message):
        if (message.__class__ in self._handlers):
            return True

        if (hasattr(service, '__partials__') and message.__class__ in service.__partials__):
            return True

        return False
        
            
    def command(self, command_type: type):
        def inner_function(func):
            
            if command_type in self._handlers:
                raise MessageSingleHandlerBroken(f'handler already defined for {command_type}')
                
            self._handlers.setdefault(command_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function

    def integration(self, integration_type: type):
        def inner_function(func):
            
            self._handlers.setdefault(integration_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function
        
    def query(self, query_type: type):
        def inner_function(func):
            
            if query_type in self._handlers:
                raise MessageSingleHandlerBroken(f'handler already defined for {query_type}')
            
            self._handlers.setdefault(query_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function
        
    def event(self, event_type: type):
        def inner_function(func):
            
            self._handlers.setdefault(event_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function

    def trace(self, *messages):
        def inner_function(func):
            def wrapper(*args, **kwargs):
                service = args[0]
                trace = kwargs.pop('__trace__')
                leadings = kwargs.pop('__leadings__')

                if(not hasattr(service, '__partials__')):
                    service.__partials__ = dict()

                if len(leadings) > 0:
                    new_trace = []
                    new_trace.extend(trace)
                    new_trace.append(args[1])

                    service.__partials__.setdefault(leadings[0], set()).add(partial(wrapper, __trace__=new_trace, __leadings__=leadings[1:]))
                else:
                    return func(service, *trace, *args[1:], **kwargs)

            self._handlers.setdefault(messages[0], set()).add(partial(wrapper, __trace__=[], __leadings__=messages[1:]))

            return wrapper

        return inner_function

    


from typing import Type
from collections import namedtuple

from domainpy.domain.model.event import DomainEvent


EventRecord = namedtuple(
    'EventRecord', 
    ('stream_id', 'number', 'topic', 'version', 'timestamp', 'trace_id', 'message', 'context', 'payload')
)


class EventMapper:
    
    def __init__(self, context):
        self.context = context

        self.map = dict()
        
    def register(self, cls: Type[DomainEvent]):
        self.map[cls.__name__] = cls
        return cls

    def is_event(self, topic: str):
        return topic in self.map
    
    def serialize(self, event: DomainEvent):
        if hasattr(event.__class__, '__annotations__'):
            attrs = event.__class__.__dict__['__annotations__']
            
            return EventRecord(
                stream_id=event.__aggregate_id__, # pylint: disable=maybe-no-member
                number=event.__number__, # pylint: disable=maybe-no-member
                topic=event.__class__.__name__,
                version=event.__version__, # pylint: disable=maybe-no-member
                timestamp=event.__timestamp__, # pylint: disable=maybe-no-member
                trace_id=event.__trace_id__,
                message=event.__message__,
                context=self.context,
                payload=event.__to_dict__()
            )
        else:
            raise NotImplementedError(
                f'{event.__class__.__name__} should have annotations'
            )
    
    def deserialize(self, event_record: EventRecord):
        event_class = self.map[event_record.topic]
        event = event_class.__from_dict__(event_record.payload)
        
        event.__dict__.update({
            '__aggregate_id__': event_record.stream_id,
            '__number__': event_record.number,
            '__version__': event_record.version,
            '__timestamp__': event_record.timestamp,
            '__trace_id__':  event_record.trace_id,
            '__message__': event_record.message
        })
        
        return event
    


import boto3
from boto3.dynamodb.conditions import Key, Attr

from domainpy.infrastructure.eventsourced.recordmanager import (
    EventRecordManager,
    Session
)
from domainpy.infrastructure.exception import (
    ConcurrencyException
)
from domainpy.utils.mappers.eventmapper import EventRecord
from domainpy.utils.dynamodb import serialize, deserialize


class DynamoEventRecordManager(EventRecordManager):

    def __init__(self, table_name, region_name=None):
        self.dynamodb = boto3.resource('dynamodb', region_name=region_name)
        self.table = self.dynamodb.Table(table_name)

    def session(self):
        return DynamoSession(self)

    def find(self, stream_id: str):
        query = self.table.query(
            KeyConditionExpression=Key('stream_id').eq(stream_id)
        )
        return tuple([
            EventRecord(
                stream_id=deserialize(i['stream_id']),
                number=deserialize(i['number']),
                topic=deserialize(i['topic']),
                version=deserialize(i['version']),
                timestamp=deserialize(i['timestamp']),
                trace_id=deserialize(i['trace_id']),
                message=deserialize(i['message']),
                context=deserialize(i['context']),
                payload=deserialize(i['payload'])
            )
            for i in query['Items']
        ])


class DynamoSession(Session):

    def __init__(self, record_manager):
        self.record_manager = record_manager

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        #self.writer.__exit__(*args, **kwargs)
        pass

    def append(self, event_record: EventRecord):
        if event_record is None:
            raise TypeError('event_record cannot be None')
        
        try:
            self.record_manager.table.put_item(
                Item={
                    'stream_id': serialize(event_record.stream_id),
                    'number': serialize(event_record.number),
                    'topic': serialize(event_record.topic),
                    'version': serialize(event_record.version),
                    'timestamp': serialize(event_record.timestamp),
                    'trace_id': serialize(event_record.trace_id),
                    'message': serialize(event_record.message),
                    'context': serialize(event_record.context),
                    'payload': serialize(event_record.payload)
                },
                ConditionExpression=(
                    Attr('stream_id').not_exists()
                    & Attr('number').not_exists()
                )
            )
        except self.record_manager.dynamodb.meta.client.exceptions.ConditionalCheckFailedException:
            stream_id = event_record.stream_id
            number = event_record.number
            raise ConcurrencyException(f'Other thread already write stream {stream_id} with number {number}')

    def commit(self):
        pass

    def rollback(self):
        pass
    


import time
from uuid import uuid4

from domainpy.utils.constructable import Constructable
from domainpy.utils.immutable import Immutable
from domainpy.utils.dictable import Dictable
from domainpy.utils.traceable import Traceable

from domainpy.application.command import ApplicationCommand


class IntegrationEvent(Constructable, Immutable, Dictable, Traceable):

    def __init__(self, *args, **kwargs):
        self.__dict__.update({
            '__trace_id__': kwargs.pop('__trace_id__', Traceable.__trace_id__),
            '__timestamp__': time.time(),
            '__message__': 'integration',
            '__error__': kwargs.pop('__error__', None)
        })
        
        super(IntegrationEvent, self).__init__(*args, **kwargs)

    @classmethod
    def from_command(cls, command: ApplicationCommand, **kwargs):
        payload = {
            k:v 
            for k,v in command.__dict__.items() 
            if not k.startswith('__')
        }
        return cls(**payload, **kwargs)

    class Resolution:
        success = 'success'
        failure = 'failure'



from setuptools import setup, find_packages


setup(
    name='domainpy',
    version='0.1.22',
    description='A library for DDD, ES, CQRS, TDD + BDD and microservices',
    author='Brian Estrada',
    author_email='brianseg014@gmail.com',
    packages=find_packages(),
    license='MIT',
    url='https://github.com/mymamachef/domainpy',
    download_url='https://github.com/mymamachef/domainpy/archive/v0.1.22.tar.gz',
    keywords=['ddd', 'event sourcing', 'CQRS'],
    install_requires=['boto3==1.17.40']
)




from typing import Callable

from domainpy.application.command import ApplicationCommand

class ApplicationService:
    
    def __handle__(self, msg):
        self.handle(msg)
        
    def handle(self, msg):
        pass
    

    


import decimal

def serialize(obj):
    if isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = serialize(obj[i])
        return obj
    elif isinstance(obj, dict):
        for k in obj.keys():
            obj[k] = serialize(obj[k])
        return obj
    elif isinstance(obj, float):
        return decimal.Decimal(str(obj))
    else:
        return obj


def deserialize(obj):
    if isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = deserialize(obj[i])
        return obj
    elif isinstance(obj, dict):
        for k in obj.keys():
            obj[k] = deserialize(obj[k])
        return obj
    elif isinstance(obj, decimal.Decimal):
        if obj % 1 == 0:
            return int(obj)
        else:
            return float(obj)
    else:
        return obj


import boto3
import json

from domainpy.utils.bus import Bus


class SimpleQueueServiceBus(Bus):

    def __init__(self, mapper, region_name=None):
        self.mapper = mapper

        self.sqs = boto3.resource('sqs', region_name=region_name)
        self.names = []

    def attach(self, handler):
        self.names.append(handler)

    def detach(self, handler):
        self.names.remove(handler)

    def publish(self, publishable):
        for name in self.names:
            queue = self.sqs.get_queue_by_name(QueueName=name)
            queue.send_message(
                MessageBody=json.dumps(self.mapper.serialize(publishable)._asdict())
            )
        



from domainpy.infrastructure.eventsourced.recordmanager import (
    EventRecordManager,
    Session
)
from domainpy.utils.mappers.eventmapper import EventRecord

class MemoryEventRecordManager(EventRecordManager):
    
    def __init__(self):
        self._heap = []
        
    def session(self):
         return MemorySession(self)
     
    def find(self, stream_id: str):
        return (
            er
            for er in self._heap
            if er.stream_id == stream_id
        )
        

class MemorySession(Session):
    
    def __init__(self, record_manager):
        self.record_manager = record_manager
        
        self._heap = []
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args, **kwargs):
        self.rollback()
    
    def append(self, event_record: EventRecord):
        if event_record is None:
            raise TypeError('event_record cannot be none')
        
        self._heap.append(event_record)
    
    def commit(self):
        self.record_manager._heap.extend(self._heap)
    
    def rollback(self):
        pass




import json
from typing import TypeVar, Generic
from datetime import datetime, date
from uuid import UUID, uuid4

from domainpy.utils.constructable import Constructable
from domainpy.utils.immutable import Immutable
from domainpy.utils.dictable import Dictable
from domainpy.domain.exceptions import ValueObjectIsNotSerializable


class ValueObject(Constructable, Immutable, Dictable):
    
    def __hash__(self):
        return hash(json.dumps(self.__to_dict__(), sort_keys=True))
    
    def __eq__(self, other):
        if other is None:
            return False
        
        return isinstance(other, ValueObject) and self.__hash__() == other.__hash__()

    def __ne__(self, other):
        return not (self == other)
    
    def __repr__(self):
        return f'{self.__class__.__name__}({json.dumps(self.__to_dict__())})'

class Identity(ValueObject):
    
    def __init__(self, id: str):
        if hasattr(self.__class__, '__annotations__'):
            attrs = self.__class__.__dict__['__annotations__']
            
            assert attrs == { "id": str }
        
        if not isinstance(id, str):
            raise TypeError(f'id should be instance of str, found {id.__class__.__name__}')
        
        self.__dict__.update(id=id)
        
    @classmethod
    def from_text(cls, id: str):
        return cls(id=id)
    
    @classmethod
    def create(cls):
        return cls.from_text(
            str(uuid4())
        )
    



import time

from domainpy.domain.model.value_object import Identity
from domainpy.domain.model.event import DomainEvent
from domainpy.utils.traceable import Traceable


class AggregateRoot(Traceable):
    
    def __init__(self, id: Identity):
        if not isinstance(id, Identity):
            raise TypeError(f'id should be type of Identity, found {id.__class__.__name__}')
        
        self.__id__ = id
        
        self.__version__ = 0
        self.__changes__ = [] # New events
        self.__seen__ = [] # Idempotent
    
    def __apply__(self, event: DomainEvent):
        self.__stamp__(event)
        self.__route__(event)
        
        self.__changes__.append(event)
    
    def __stamp__(self, event: DomainEvent):
        event.__dict__.update({
            '__aggregate_id__': f'{self.__id__.id}:{self.__class__.__name__}',
            '__number__': self.__version__ + 1,
            '__version__': 1,
            '__timestamp__': time.time()
        })
        
    def __route__(self, event: DomainEvent):
        if event not in self.__seen__:
            self.__version__ = event.__number__
            self.__seen__.append(event)

            self.mutate(event)

    def __get_by_trace_id__(self, trace_id: str):
        return tuple([
            e for e in self.__seen__
            if e.__trace_id__ == trace_id
        ])

    def __get_for_compensation__(self, trace_id, event_type: type, compensate_type: type = None):
        events = self.__get_by_trace_id__(trace_id)
        filtered_events = tuple([
            e for e in events if isinstance(e, event_type)
        ])

        for e in events:
            print([e, e.__trace_id__])

        if compensate_type is None:
            return filtered_events
        else:
            return tuple([
                e0 for e0 in events
                if not any(
                    True for e1 in events 
                    if isinstance(e1, compensate_type)
                    and e1.__trace_id__ == e0.__trace_id__
                )
            ])
    
    def mutate(self, event: DomainEvent):
        pass
